
{-
Copyright 2009, Jitesh Jain, Shashank Kapoor, Sajal Kumar Sinha, Piyush Srivastava and Utkarsh Upadhyay.

This file is part of atp, an automated theorem prover.

atp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

atp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with atp.  If not, see <http://www.gnu.org/licenses/>.

-}



module Resolution1 where
import Logic1
import Unification1 as Unification
import Data.Map as Map
import Data.List as List
import Control.Monad.State
import Data.Sequence as Seq


data LStateInt = LStateInt {lSLowestInt::Int, lSTableInt::Map.Map Int Int}
type States a = State LStateInt a

{- | For variable substitution, assuming all variables are Var
Wrapper function-}
modify :: Int -> CnfFormula -> (Int, CnfFormula)
modify n f = let (c, LStateInt {lSLowestInt=a, lSTableInt=b}) = runState (modify' f) LStateInt{lSLowestInt = n, lSTableInt = Map.empty} 
               in (a, c)

{- |This function takes a cnfFormula, and depending upon current state,
     substitutes the variables to be unique. -}
modify' :: CnfFormula -> States CnfFormula
modify' = mapM modifyLiteral
--modify' [] = return []
--modify' (x:xs) = do a <- modifyLiteral a
--                    b <- modify' xs
--                    return (a:b)
--

{- | This function takes a Literal, and depending upon current state,
     substitutes the variables to be unique. -}
modifyLiteral :: Literal -> States Literal
modifyLiteral x@(Not a) = liftM Not (modifyAtomicFormula a)
modifyLiteral x@(Pos a) = liftM Pos (modifyAtomicFormula a)

{- | This function takes an atomic formula , and depending upon current state,
     substitutes the variables to be unique. -}
modifyAtomicFormula :: AtomicFormula -> States AtomicFormula
modifyAtomicFormula AtomicFormula{afName=afName', afArg=afArg'} = do {x <- mapM modifyTerm afArg'
                                                        ; return (AtomicFormula {afName=afName', afArg=x})}

{- | This function takes a term, and depending upon current state,
     substitutes the variables to be unique. -}
modifyTerm :: Term -> States Term
modifyTerm (Function {tName = tName', tArg =tArg'}) = do x <- mapM modifyTerm tArg'
                                                         return (Function{tName = tName', tArg = x})

{- | This function takes a variable, and depending upon current state,
     substitutes the variables to be unique. -}
modifyTerm (Var i) = do    state <- get
                           let low = lSLowestInt state
                               table = lSTableInt state
                               query = member i table
                               in if query 
                                     then return (Var (table ! i))
                                     else do put LStateInt{lSLowestInt = low + 1, lSTableInt = Map.insert i low table}
                                             return (Var low)


{- | This function takes a pair of formulae, and
     substitutes the variables to be unique. -}
modifyPair :: CnfFormula -> CnfFormula -> (CnfFormula, CnfFormula)
modifyPair f1 f2 = let (low, form1) = Resolution1.modify 0 f1
                       form2 = snd $ Resolution1.modify low f2
                       in (form1, form2)





--SUBSTITUTONS
type Subst = Binding
--The first term is always a Var


--Apply a list of Substitutions one by one

substitute :: [Subst] -> CnfFormula -> CnfFormula
substitute l f = foldl (flip substituteCnf) f l 

substituteCnf :: Subst -> CnfFormula -> CnfFormula
substituteCnf a = Prelude.map (substituteLiteral a)

substituteLiteral :: Subst -> Literal -> Literal
substituteLiteral a (Not x) = Not (substituteAF a x)
substituteLiteral a (Pos x) = Pos (substituteAF a x)

substituteAF :: Subst -> AtomicFormula -> AtomicFormula
substituteAF a (AtomicFormula{afName = name, afArg = arg}) = AtomicFormula{afName = name, 
                                                                afArg = Prelude.map (substituteTerm a) arg}

substituteTerm :: Subst -> Term -> Term                                                                               
substituteTerm a (Function{tName = name, tArg = arg}) = Function{tName = name, 
                                                                tArg = Prelude.map (substituteTerm a) arg}
substituteTerm (v, t) (Var u) | u == v  = t 
substituteTerm _ a  = a

type IndexedClause = (Int, CnfFormula) 

{- |Finding if the precondition for ordered resolution is satisfied-}

startsNot :: LowerWord -> Literal -> Bool
startsNot s (Not a) = s == afName a
startsNot _ _ = False


startsPos :: LowerWord -> Literal -> Bool
startsPos s (Pos a) = s == afName a
startsPos _ _ = False

stripAF :: Literal -> AtomicFormula
stripAF (Not a) = a
stripAF (Pos a) = a



merge :: (Ord a) => [a] -> [a] -> [a]
merge [] a = a
merge a [] = a
merge s@(x:xs) t@(y:ys) = if (x < y) then (x:(merge xs t)) else (y:(merge s ys)) 

litName :: Literal -> LowerWord
litName (Not a) = afName a
litName (Pos a) = afName a

getNots:: Literal -> Bool
getNots (Not _) =  True
getNots _ = False
  

{- | The function to be called by the resolver, for getting the list of
- Atomicformulas. Returns Nothing if resolution is apriori not possible. -}

isInOrderForSup :: CnfFormula -> CnfFormula -> [(Literal, Literal)]
isInOrderForSup p q = let       (pN, pP) = List.partition getNots p
                                (qN, qP) = List.partition getNots q
                                in do   x <- pN
                                        y <- qP
                                        if (litName x == litName y)
                                           then return (x,y)
                                           else mzero
                                   `mplus` do   x <- pP
                                                y <- qN
                                                if (litName x == litName y)
                                                   then return (x,y)
                                                   else mzero


