
{-
Copyright 2009, Jitesh Jain, Shashank Kapoor, Sajal Kumar Sinha, Piyush Srivastava and Utkarsh Upadhyay.

This file is part of atp, an automated theorem prover.

atp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

atp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with atp.  If not, see <http://www.gnu.org/licenses/>.

-}


module Unification1 where
import Logic1
import Data.List

type UnificationElement = Term
                  -- TODO: Is it an overkill already?

type Binding          = (UpperWord, Term)
                  -- Our binding is a Variable names assigned to a Term
                  
type Bindings          = Either [Binding] Error
                  -- This is our environment.

type Error          = String
                  -- This is the error string.

unify :: UnificationElement -> UnificationElement -> Bindings -> Bindings
unify _ _ r@(Right err) = r
unify x y bindings = if x == y 
			then bindings
			else case (x,y) of
                                   (Var a, _) -> unifyVar a y bindings
                                   (_, Var a) -> unifyVar a x bindings
                                   (Function n1 arg1, Function n2 arg2) | n1==n2 -> unifyList arg1 arg2 bindings 
                                   otherwise -> Right ("Cannot unify: " ++ (show x) ++ " and " ++ (show y))

unifyList :: [UnificationElement] -> [UnificationElement] -> Bindings -> Bindings
unifyList _ _ r@(Right err) = r
unifyList x y bindings = 
         case (x,y) of
           ([], []) -> bindings
           (x:xs, y:ys) -> unifyList xs ys (unify x y bindings)
           otherwise -> Right ("Cannot unify arguments of unequal size.")

unifyVar :: UpperWord -> UnificationElement -> Bindings -> Bindings
unifyVar _ _ r@(Right err) = r
unifyVar var term bindings@(Left env) =
         case find (variable var) env of
           Just (_, value) -> unify value term bindings
           Nothing         -> case term of
                                      Var up -> case find (variable up) env of
                                                 Just (_, value) -> unify (Var var) value bindings
                                                 Nothing -> finally ()
                                      otherwise -> finally ()
         where    variable var x = (fst x) == var
                  finally () = if occursCheck var term 
                                    then Right ("Occurs Check failed: " ++ show var ++ " in " ++ show(term))
                                         else Left ((var, term):env)


occursCheck :: UpperWord -> Term -> Bool
occursCheck var (Var _) = False
occursCheck var term = case (find (==var) $ getVariables term) of
                            Just _ -> True
                            Nothing -> False
                               
performSubstutions :: Bindings -> Bindings
performSubstutions r@(Right err) = r
performSubstutions (Left bindings) = Left (performSubstutions_ 0 bindings)



performSubstutions_ :: Int -> [Binding] -> [Binding]
performSubstutions_ idx bindings =
         if length bindings == idx then bindings
         else let (var, value) = bindings !! idx in
                  performSubstutions_ (idx+1) (map (alter var value) bindings)
                       where alter var value (v, term) = if (v==var) then (v, term) else (v, substitute var value term)

substitute :: UpperWord -> Term -> Term -> Term
substitute var substitutionValue term =
         case term of
           Var v | v == var -> substitutionValue
                    | otherwise -> Var v
           Function n args  -> Function {tName=n, tArg=(map (substitute var substitutionValue) args)}

unifyTogether :: [Term] -> Bindings
unifyTogether [] = Left []
unifyTogether (x:xs) = unifyTogether_ x xs (Left [])

unifyTogether_ :: Term -> [Term] -> Bindings -> Bindings
unifyTogether_ _ _ err@(Right _) = err
unifyTogether_ tr [] bindings          = bindings
unifyTogether_ tr (x:xs) bindings  = unifyTogether_ x xs (unify x tr bindings)

--A version of UnifyList to be able to handle a list of Atomic Formulae
--It would be given a list of atomic formulae with the same signature; but
--will nonetheless check for matching argument list size.


afToListOfTerms :: [AtomicFormula] -> [[Term]]
afToListOfTerms s@(p:l) = let a = length (afArg p)
                              temp :: Int -> AtomicFormula -> (Int, [Term])
                              temp n f = let s = afArg f
                                             in (max n (length s), s)
                              (n, list) = mapAccumL temp a s 
                              in if n /= a 
                                    then error $ "Fatal error: Predicates with different arg sizes" ++ show s
                                    else transpose list





unifyAfs_ :: [[Term]] -> Bindings -> Bindings
unifyAfs_ _ r@(Right err) = r
unifyAfs_ l bindings = 
         case (l) of
           ([]) -> bindings
           (x:xs) -> unifyAfs_ xs (unifyAfsTogether_ x bindings)

unifyAfsTogether_ :: [Term] -> Bindings -> Bindings
unifyAfsTogether_ _ err@(Right _) = err
unifyAfsTogether_ [] bindings          = bindings
unifyAfsTogether_ (x:ys) bindings = unifyTogether_ x ys bindings
--unifyAfsTogether_ (x:y:xs) bindings  = unifyTogether_ y xs (unify y x bindings)


{- | The function to be called by the resolver, for unifying the lists of
- arguments obtained through Resolution.isInOrder. -}
unifyAFs :: [AtomicFormula] -> Bindings
unifyAFs l = let p = unifyAfs_ (afToListOfTerms l) (Left [])
                 in case p of 
                        Right _ -> p
                        Left a -> let n = (performSubstutions_ 0 a)
                                      m = foldl (||) False (map (uncurry occursCheck) n)
                                      in if m 
                                               then Right "Occurs check failed"
                                               else Left n
