module Unification_map where
import Logic
import Data.List as List
import Data.Map as Map

type UnificationElement = Term
                -- TODO: Is it an overkill already?

type Binding         = (UpperWord, Term)
                -- Our binding is a Variable names assigned to a Term

type InternalBinding = Map UpperWord Term
                
type Bindings         = Either [Binding] Error
                -- This is our environment.

type InternalBindings = Either InternalBinding Error

type Error         = String
                -- This is the error string.

unify :: UnificationElement -> UnificationElement -> InternalBindings -> InternalBindings
unify _ _ r@(Right err) = r
unify x y bindings = 
        if x == y 
           then bindings
           else case (x,y) of
                  (Var a, _) -> unifyVar a y bindings
                  (_, Var a) -> unifyVar a x bindings
                  (Function n1 arg1, Function n2 arg2) | n1==n2 -> unifyList arg1 arg2 bindings 
                  otherwise -> Right ("Cannot unify: " ++ (show x) ++ " and " ++ (show y))

unifyList :: [UnificationElement] -> [UnificationElement] -> InternalBindings -> InternalBindings
unifyList _ _ r@(Right err) = r
unifyList x y bindings = 
	case (x,y) of
	  ([], []) -> bindings
	  (x:xs, y:ys) -> unifyList xs ys (unify x y bindings)
	  otherwise -> Right ("Cannot unify arguments of unequal size.")

unifyVar :: UpperWord -> UnificationElement -> InternalBindings -> InternalBindings
unifyVar _ _ r@(Right err) = r
unifyVar var term bindings@(Left env) =
	case Map.lookup var env of
	  Just value 	  -> unify value term bindings
	  Nothing 	  -> case term of
	  			Var up -> case Map.lookup up env of
					    Just value 	-> unify (Var var) value bindings
					    Nothing 	-> finally ()
				otherwise -> finally ()
	where 	finally () = if occursCheck var term 
				then Right ("Occurs Check failed: " ++ show var ++ " in " ++ show(term))
			     	else Left (Map.insert var term env)


occursCheck :: UpperWord -> Term -> Bool
occursCheck var (Var _) = False
occursCheck var term = case (find (==var) $ getVariables term) of
			 Just _ -> True
			 Nothing -> False
	  		  
performSubstutions :: Bindings -> Bindings
performSubstutions r@(Right err) = r
performSubstutions (Left bindings) = Left (performSubstutions_ 0 bindings)

performSubstutions_ :: Int -> [Binding] -> [Binding]
performSubstutions_ idx bindings =
	if length bindings == idx then bindings
	else let (var, value) = bindings !! idx in
		performSubstutions_ (idx+1) (List.map (alter var value) bindings)
	     	where alter var value (v, term) = if (v==var) then (v, term) else (v, substitute var value term)

substitute :: UpperWord -> Term -> Term -> Term
substitute var substitutionValue term =
	case term of
	  Var v | v == var -> substitutionValue
	  	| otherwise -> Var v
	  Function n args  -> Function {tName=n, tArg=(List.map (substitute var substitutionValue) args)}

unifyTogether :: [Term] -> Bindings
unifyTogether [] = Left []
unifyTogether (x:xs) = case unifyTogether_ x xs (Left Map.empty) of
		  	Right err -> Right err
			Left map  -> Left (toList map)

unifyTogether_ :: Term -> [Term] -> InternalBindings -> InternalBindings
unifyTogether_ _ _ err@(Right _) = err
unifyTogether_ tr [] bindings	 = bindings
unifyTogether_ tr (x:xs) bindings  = unifyTogether_ x xs (unify x tr bindings)

--A version of UnifyList to be able to handle a list of Atomic Formulae
--It would be given a list of atomic formulae with the same signature; but
--will nonetheless check for matching argument list size.


afToListOfTerms :: [AtomicFormula] -> [[Term]]
afToListOfTerms s@(p:l) = let a = length (afArg p)
                              temp :: Int -> AtomicFormula -> (Int, [Term])
                              temp n f = let s = afArg f
                                             in (max n (length s), s)
                              (n, list) = mapAccumL temp a s 
                              in if n /= a 
                                    then error "Fatal error: Predicates with different arg sizes"
                                    else transpose list



-- The functions that unify Atomic Formula

unifyAfs_ :: [[Term]] -> InternalBindings -> InternalBindings
unifyAfs_ _ r@(Right err) = r
unifyAfs_ l bindings = 
	case (l) of
	  ([]) -> bindings
	  (x:xs) -> unifyAfs_ xs (unifyAfsTogether_ x bindings)

unifyAfsTogether_ :: [Term] -> InternalBindings -> InternalBindings
unifyAfsTogether_ _ err@(Right _) = err
unifyAfsTogether_ [] bindings	 = bindings
unifyAfsTogether_ (x:ys) bindings = unifyTogether_ x ys bindings
--unifyAfsTogether_ (x:y:xs) bindings  = unifyTogether_ y xs (unify y x bindings)


{- | The function to be called by the resolver, for unifying the lists of
- arguments obtained through Resolution.isInOrder. -}
unifyAFs :: [AtomicFormula] -> Bindings
unifyAFs l = let p = unifyAfs_ (afToListOfTerms l) (Left Map.empty)
                 in case p of 
                        Right err -> Right err
                        Left a -> let n = (performSubstutions_ 0 (toList a))
                                      m = foldl (||) False (List.map (uncurry occursCheck) n)
                                      in if m 
                                               then Right "Occurs check failed"
                                               else Left n
